@@ -1,4 +1,4 @@
-/* $Id: mode.c,v 1.11 2002/12/18 08:09:22 cegger Exp $
+/* $Id$
 ******************************************************************************
 
    Graphics library for GGI. X target.
@@ -29,15 +29,15 @@
 ******************************************************************************
 */
 
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
-
 #include "config.h"
 #include <ggi/internal/ggi-dl.h>
 #include <ggi/display/x.h>
 #include <ggi/display/mansync.h>
 
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+
 
 /*
  * "this" is a C++ keyword, so we use "cthis" instead.
@@ -56,7 +56,7 @@ static int GGI_X_checkmode_internal(ggi_visual *vis, ggi_mode *tm, int *viidx)
 
 	LIBGGI_APPASSERT(vis != NULL, "GGIcheckmode: vis == NULL");
 
-	priv = LIBGGI_PRIVATE(vis);
+	priv = GGIX_PRIV(vis);
 
 	best = NULL;
 	w = h = idx = 0;
@@ -76,7 +76,7 @@ static int GGI_X_checkmode_internal(ggi_visual *vis, ggi_mode *tm, int *viidx)
 		if (_ggi_x_fit_geometry(vis, tm, cthis, &dummy) != GGI_OK) 
 			continue;
 
-		if ((dummy.visible.x * dummy.visible.y) < (w * h))
+		if ((unsigned)(dummy.visible.x * dummy.visible.y) < (w * h))
 			continue;
 
 		best = cthis;
@@ -99,7 +99,7 @@ static int GGI_X_checkmode_internal(ggi_visual *vis, ggi_mode *tm, int *viidx)
 	best = NULL;
 	w = h = idx = 0;
 	while (idx < priv->nvisuals) {
-	  	ggi_x_vi	*cthis;
+	  ggi_x_vi	*cthis;
 		ggi_mode	dummy;
 
 		dummy.visible.x = dummy.visible.y = 0;
@@ -113,7 +113,7 @@ static int GGI_X_checkmode_internal(ggi_visual *vis, ggi_mode *tm, int *viidx)
 			continue;
 		_ggi_x_fit_geometry(vis, tm, cthis, &dummy);
 
-		if ((dummy.visible.x * dummy.visible.y) < (w * h))
+		if ((unsigned)(dummy.visible.x * dummy.visible.y) < (w * h))
 			continue;
 
 		best = cthis;
@@ -128,7 +128,7 @@ static int GGI_X_checkmode_internal(ggi_visual *vis, ggi_mode *tm, int *viidx)
 	if (_ggi_x_fit_geometry(vis, tm, best, tm) != GGI_OK) {
 		GGIDPRINT("This should not happen\n");
 	}
-	return GGI_OK;
+	return -1;
 
  suggest2:
 	/* Adjust the graphtype up one notch if possible.  
@@ -184,39 +184,86 @@ static int GGI_X_checkmode_internal(ggi_visual *vis, ggi_mode *tm, int *viidx)
 	return -1;
 }
 
-int GGI_X_checkmode_normal(ggi_visual *vis, ggi_mode *tm) {
-	int dummy;
-	return(GGI_X_checkmode_internal(vis, tm, &dummy));
-}
-
-int GGI_X_checkmode_fixed(ggi_visual *vis, ggi_mode *tm) {
+int GGI_X_checkmode_normal(ggi_visual *vis, ggi_mode *tm)
+{
 	ggi_x_priv *priv;
-	Window root;
-	int w, h, dummy;
+	int dummy;
+	int rc;
+	rc = GGI_X_checkmode_internal(vis, tm, &dummy);
 
 	priv = GGIX_PRIV(vis);
 
-#warning retcode checking here
-        XGetGeometry(priv->disp, priv->parentwin, &root, &dummy, &dummy,
-		     &w, &h, &dummy, &dummy);
+	GGIDPRINT_MODE("X (checkmode_normal): mlfuncs.validate = %p\n",
+			priv->mlfuncs.validate);
+	if (priv->mlfuncs.validate != NULL) {
+		priv->cur_mode = priv->mlfuncs.validate(vis, -1, tm);
+		if (priv->cur_mode < 0) {
+			GGIDPRINT_MODE("X: mlfuncs.validate failed: %i\n",
+				priv->cur_mode);
+			/* An error occured */
+			dummy = priv->cur_mode;
+			priv->cur_mode = 0;
+			return dummy;
+		}	/* if */
+		GGIDPRINT_MODE("X: mlfuncs.validate successful: %i\n",
+			priv->cur_mode);
+	}	/* if */
+
+	return rc;
+}
+
+int GGI_X_checkmode_fixed(ggi_visual *vis, ggi_mode *tm)
+{
+	ggi_x_priv *priv;
+	Window root;
+	int w, h;
+	unsigned int depth, dummy;
+
+	priv = GGIX_PRIV(vis);
+
+	if (! XGetGeometry(priv->disp, priv->parentwin, &root,
+		     (int *)&dummy, (int *)&dummy,
+		     &w, &h, &dummy, &depth))
+	{
+		GGIDPRINT_MODE("X (checkmode_fixed):"
+				"no reply from X11 server\n");
+		return GGI_EEVNOTARGET;
+	}
 
 	if (tm->visible.x == GGI_AUTO) tm->visible.x = w;
 	if (tm->visible.y == GGI_AUTO) tm->visible.y = h;
 	if ((tm->visible.x != w) || (tm->visible.y != h))
 		return GGI_EARGINVAL;
 
-	dummy = GGI_X_checkmode_internal(vis, tm, &dummy);
+	dummy = GGI_X_checkmode_internal(vis, tm, (int *)&dummy);
 	if ((dummy != GGI_OK) || (tm->visible.x != w) || (tm->visible.y != h)) {
 		tm->visible.x = w;
 		tm->visible.y = h;
 	}
+
+	GGIDPRINT_MODE("X (checkmode_fixed): mlfuncs.validate = %p\n",
+			priv->mlfuncs.validate);
+	if (priv->mlfuncs.validate != NULL) {
+		priv->cur_mode = priv->mlfuncs.validate(vis, -1, tm);
+		if (priv->cur_mode < 0) {
+			GGIDPRINT_MODE("X: mlfuncs.validate failed: %i\n",
+					priv->cur_mode);
+			/* An error occured */
+			dummy = priv->cur_mode;
+			priv->cur_mode = 0;
+			return dummy;
+		}	/* if */
+		GGIDPRINT_MODE("X: mlfuncs.validate successful: %i\n",
+				priv->cur_mode);
+	}	/* if */
+
 	return dummy;
 }
 
 static int ggi_x_load_mode_libs(ggi_visual *vis)
 {
 	int err,id;
-	char sugname[256],args[256];
+	char sugname[GGI_MAX_APILEN],args[GGI_MAX_APILEN];
 
 	_ggiZapMode(vis, 0);
         for (id=1; 0 == vis->opdisplay->getapi(vis,id,sugname,args); id++) {
@@ -276,13 +323,17 @@ int GGI_X_setmode_normal(ggi_visual *vis, ggi_mode *tm)
 	ggi_x_priv      *priv;
 	int destroychild, destroyparent, createchild, createparent;
 
+	GGIDPRINT("GGI_X_setmode_normal(%p, %p) called\n",vis,tm);
+
 	priv = GGIX_PRIV(vis);
 
 	if ((err=GGI_X_checkmode_internal(vis,tm,&viidx))) return err;
 	memcpy(LIBGGI_MODE(vis), tm, sizeof(ggi_mode));
 	priv->viidx = viidx;
 
-	GGIDPRINT("viidx = %i\n", priv->viidx);
+	GGIDPRINT("* viidx = %i\n", priv->viidx);
+
+	if (priv->opmansync) MANSYNC_ignore(vis);
 
 	ggLock(priv->xliblock);
 
@@ -290,6 +341,16 @@ int GGI_X_setmode_normal(ggi_visual *vis, ggi_mode *tm)
 
 	_ggi_x_build_pixfmt(vis, tm, vi); /* Fill in ggi_pixelformat */
 
+	/* Restore original mode */
+	GGIDPRINT_MODE("X (setmode_normal): mlfuncs.restore = %p\n",
+			priv->mlfuncs.restore);
+	if (priv->mlfuncs.restore != NULL) {
+		err = priv->mlfuncs.restore(vis);
+		GGIDPRINT_MODE("X: mlfuncs.restore retcode: %i\n",
+				err);
+		if (err) goto err0;
+	}	/* if */
+	
 #warning TODO: do not destroy windows unless necessary.
 
 	destroychild = destroyparent = 1;
@@ -314,45 +375,66 @@ int GGI_X_setmode_normal(ggi_visual *vis, ggi_mode *tm)
 	priv->parentwin = 
 		XCreateSimpleWindow(priv->disp,
 				    RootWindow(priv->disp, vi->screen),
-				    0, 0, tm->visible.x, tm->visible.y, 0,
+				    0, 0,
+				    (unsigned int)tm->visible.x,
+				    (unsigned int)tm->visible.y, 0,
 				    0, 0);
 	_ggi_x_dress_parentwin(vis, tm);
 
 	GGIDPRINT_MODE("X: Prepare to resize.\n");
-	XResizeWindow(priv->disp,priv->parentwin,tm->visible.x,tm->visible.y);
-	GGIDPRINT_MODE("X: About to map parent\n");
+	XResizeWindow(priv->disp,priv->parentwin,
+			(unsigned int)tm->visible.x,
+			(unsigned int)tm->visible.y);
+	GGIDPRINT_MODE("X: About to map parent (%p)\n", priv->parentwin);
+
 	/* Map window. */
+	GGIDPRINT_MODE("X: Parent win: Map Input\n");
 	XSelectInput(priv->disp, priv->parentwin, ExposureMask);
+	GGIDPRINT_MODE("X: Parent win: Raise Mapping\n");
 	XMapRaised(priv->disp, priv->parentwin);
-	
+	GGIDPRINT_MODE("X: Parent win: Map requested\n");
+
 	/* Wait for window to become mapped */
 	XNextEvent (priv->disp, &event);
 	GGIDPRINT_MODE("X: Window Mapped\n");
-	
-	/* Parent window doesn't listen for input. */
-	XSelectInput(priv->disp, priv->parentwin, 0);
+
+	/* We let the parent window listen for the keyboard.
+	 * this allows to have the windowmanager decide about keyboard 
+	 * focus as usual.
+	 * The child window listens for the rest.
+	 * Note, that the child window also listens for the keyboard for
+	 * those cases where we don't have a parent.
+	 */
+	XSelectInput(priv->disp, priv->parentwin,
+		     KeymapStateMask | KeyPressMask | KeyReleaseMask);
 
 oldparent:
 	GGIDPRINT_MODE("X: running in parent window 0x%x\n", priv->parentwin);
 
 	ggi_x_load_mode_libs(vis);
 	_ggi_x_load_slaveops(vis);
-	GGIDPRINT("viidx = %i\n", priv->viidx);
+	GGIDPRINT("* viidx = %i\n", priv->viidx);
 	if (priv->createfb != NULL) {
 		err = priv->createfb(vis);
-		if (err) goto err0;
+		if (err) {
+			/* xlib lock is still acquired here - unlock before exiting */
+			GGIDPRINT("priv->createfb failed.\n");
+			ggUnlock(priv->xliblock);
+			goto err0;
+		}
 	}
 
 	_ggi_x_free_colormaps(vis);
 	XSync(priv->disp, 0);
+	GGIDPRINT("Create colormap.\n");
 	_ggi_x_create_colormaps(vis, vi);
 
 	/* Create the child window */
 	attrib.colormap = priv->cmap;
 	attrib.border_pixel = BlackPixel(priv->disp, vi->screen);
 	priv->win = XCreateWindow(priv->disp, priv->parentwin,
-				  0, 0, tm->virt.x, 
-				  tm->virt.y * (tm->frames + 1), 0,
+				  0, 0, (unsigned)tm->virt.x, 
+				  (unsigned)tm->virt.y * tm->frames, 0,
 				  vi->depth, InputOutput,
 				  vi->visual, CWColormap | CWBorderPixel, 
 				  &attrib);
@@ -364,14 +446,14 @@ oldparent:
 	/* Map window. */
 	XSelectInput(priv->disp, priv->win, ExposureMask);
 	XMapWindow(priv->disp, priv->win);
-	
+
 	/* Wait for window to become mapped */
 	XNextEvent (priv->disp, &event);
 	GGIDPRINT_MODE("X: Window Mapped\n");
-	
+
 	/* Select input events to listen for */
 	XSelectInput(priv->disp, priv->win,
-		     KeyPressMask | KeyReleaseMask |
+		     KeymapStateMask | KeyPressMask | KeyReleaseMask |
 		     ButtonPressMask | ButtonReleaseMask |
 		     EnterWindowMask | LeaveWindowMask |
 		     ExposureMask | PointerMotionMask);
@@ -387,8 +469,9 @@ oldparent:
 	XSetGraphicsExposures(priv->disp, priv->tempgc, True);
 	if (priv->textfont)
 		XSetFont(priv->disp, priv->tempgc, priv->textfont->fid);
-	_ggi_x_set_xclip(priv->disp, priv->tempgc, 0, 0, LIBGGI_VIRTX(vis), 
-			 LIBGGI_VIRTY(vis) * (vis->mode->frames + 1));
+	_ggi_x_set_xclip(NULL, priv->disp, priv->tempgc, 0, 0, 
+			 LIBGGI_VIRTX(vis), 
+			 LIBGGI_VIRTY(vis) * LIBGGI_MODE(vis)->frames);
 	GGIDPRINT_MODE("X GCs allocated.\n");
 
 	/* Create a cursor (frees old cursor) */
@@ -409,32 +492,56 @@ oldparent:
 		if (err) goto err1;
 	}
 
+	GGIDPRINT_MODE("X (setmode_normal): mlfuncs.enter = %p\n",
+			priv->mlfuncs.enter);
+	if (priv->mlfuncs.enter != NULL) {
+		err = priv->mlfuncs.enter(vis, priv->cur_mode);
+		GGIDPRINT_MODE("X: mlfuncs.enter retcode = %i\n",
+				err);
+		if (err) goto err1;
+	}	/* if */
+
 	/* Tell inputlib about the new window */
 	if (priv->inp) {
 		gii_event ev;
-		gii_xwin_cmddata_setparam *data
-			= (gii_xwin_cmddata_setparam *) ev.cmd.data;
+		gii_xwin_cmddata_setparam data;
+
+		GGIDPRINT_MODE("X (setmode_normal): tell inputlib about new window\n");
+
 		ev.cmd.size = sizeof(gii_cmd_event);
 		ev.cmd.type = evCommand;
 		ev.cmd.target = priv->inp->origin;
 		ev.cmd.code = GII_CMDCODE_XWINSETPARAM;
-		data->win = priv->win;
-		if (data->win == None) data->win = priv->parentwin;
-		data->ptralwaysrel = 0;
+		data.win = priv->win;
+		if (data.win == None) {
+			data.win = priv->parentwin;
+		}
+		data.ptralwaysrel = 0;
+		data.parentwin = priv->parentwin;
+
+		/* Assure aligned memory access. Some platforms
+		 * (i.e. NetBSD/sparc64) rely on this.
+		 */
+		memcpy(ev.cmd.data, &data, sizeof(gii_xwin_cmddata_setparam));
 
 		giiEventSend(priv->inp, &ev);
 	}
 
+
+	GGIDPRINT_MODE("X (setmode_normal): set dirty region\n");
+
 	/* ggiOpen will dirty the whole screen for us by calling fillscreen */
 	priv->dirtytl.x = 1; priv->dirtybr.x = 0;
 
 	if (priv->opmansync) MANSYNC_cont(vis);
 
+	GGIDPRINT_MODE("X (setmode_normal): return code = %i\n", err);
 	return err;
 
 err1:
 	priv->freefb(vis);
 err0:
+	if (priv->opmansync) MANSYNC_cont(vis);
 	return err;
 }
 
@@ -450,6 +557,8 @@ int GGI_X_setmode_fixed(ggi_visual *vis, ggi_mode *tm)
 	XVisualInfo	*vi;
 	ggi_x_priv      *priv;
 
+	GGIDPRINT("GGI_X_setmode_fixed(%p, %p) called\n",vis,tm);
+		
 	priv = GGIX_PRIV(vis);
 
         XGetGeometry(priv->disp, priv->parentwin, &root, &dummy, &dummy,
@@ -470,6 +579,8 @@ int GGI_X_setmode_fixed(ggi_visual *vis, ggi_mode *tm)
 	memcpy(LIBGGI_MODE(vis), tm, sizeof(ggi_mode));
 	priv->viidx = viidx;
 
+	if (priv->opmansync) MANSYNC_ignore(vis);
+
 	ggLock(priv->xliblock);
 
 	vi = (priv->vilist + viidx)->vi;
@@ -486,7 +597,12 @@ int GGI_X_setmode_fixed(ggi_visual *vis, ggi_mode *tm)
 	_ggi_x_load_slaveops(vis);
 	if (priv->createfb != NULL) {
 		err = priv->createfb(vis);
-		if (err) goto err0;
+		if (err) {
+			/* xlib lock is still acquired here - unlock before exiting */
+			GGIDPRINT("priv->createfb failed.\n");
+			ggUnlock(priv->xliblock);
+			goto err0;
+		}
 	}
 
 	_ggi_x_free_colormaps(vis);
@@ -496,30 +612,40 @@ int GGI_X_setmode_fixed(ggi_visual *vis, ggi_mode *tm)
 	attrib.colormap = priv->cmap;
 	attribmask = CWBackingStore;
 	if (priv->win == root) {
+		GGIDPRINT_MODE("X (setmode_fixed): mlfuncs.restore = %p\n",
+			priv->mlfuncs.restore);
+		if (priv->mlfuncs.restore != NULL) {
+			err = priv->mlfuncs.restore(vis);
+			GGIDPRINT_MODE("X: mlfuncs.restore retcode = %i\n",
+					err);
+			if (err) goto err0;
+		}	/* if */
+
 		attribmask = CWColormap;
 		goto nochild;
 	}
 	/* Create the child window */
 	priv->win = XCreateWindow(priv->disp, priv->parentwin,
-				  0, 0, tm->virt.x, tm->virt.y, 0,
+				  0, 0, (unsigned)tm->virt.x,
+				  (unsigned)tm->virt.y, 0,
 				  vi->depth, InputOutput,
 				  vi->visual, CWColormap, &attrib);
 	GGIDPRINT_MODE("X: About to map child\n");
-	
+
 	/* Have the parent window tell the WM its children have colormaps */
 	XSetWMColormapWindows(priv->disp, priv->parentwin, &(priv->win), 1);
 
 	/* Map window. */
 	XSelectInput(priv->disp, priv->win, ExposureMask);
 	XMapWindow(priv->disp, priv->win);
-	
+
 	/* Wait for window to become mapped */
 	XNextEvent (priv->disp, &event);
 	GGIDPRINT_MODE("X: Window Mapped\n");
-	
+
 	/* Select input events to listen for */
 	XSelectInput(priv->disp, priv->win,
-		     KeyPressMask | KeyReleaseMask |
+		     KeymapStateMask | KeyPressMask | KeyReleaseMask |
 		     ButtonPressMask | ButtonReleaseMask |
 		     EnterWindowMask | LeaveWindowMask |
 		     ExposureMask | PointerMotionMask);
@@ -536,8 +662,9 @@ int GGI_X_setmode_fixed(ggi_visual *vis, ggi_mode *tm)
 	if (priv->tempgc) XFreeGC(priv->disp, priv->tempgc);
 	priv->tempgc = XCreateGC(priv->disp, priv->win, 0, 0);
 	XSetGraphicsExposures(priv->disp, priv->tempgc, True);
-	_ggi_x_set_xclip(priv->disp, priv->tempgc, 0, 0, LIBGGI_VIRTX(vis), 
-			 LIBGGI_VIRTY(vis) * (vis->mode->frames + 1));
+	_ggi_x_set_xclip(NULL, priv->disp, priv->tempgc, 0, 0, 
+			 LIBGGI_VIRTX(vis), 
+			 LIBGGI_VIRTY(vis) * LIBGGI_MODE(vis)->frames);
 	GGIDPRINT_MODE("X GCs allocated.\n");
 
 	/* Create a cursor (destroys old one) */
@@ -558,32 +685,55 @@ int GGI_X_setmode_fixed(ggi_visual *vis, ggi_mode *tm)
 		if (err) goto err1;
 	}
 
+	GGIDPRINT_MODE("X (setmode_fixed): mlfuncs.enter = %p\n",
+		priv->mlfuncs.enter);
+	if (priv->mlfuncs.enter != NULL) {
+		err = priv->mlfuncs.enter(vis, priv->cur_mode);
+		GGIDPRINT_MODE("X: mlfuncs.enter retcode = %i\n",
+				err);
+		if (err) goto err1;
+	}	/* if */
+
 	/* Tell inputlib about the new window */
 	if (priv->inp) {
 		gii_event ev;
-		gii_xwin_cmddata_setparam *data
-			= (gii_xwin_cmddata_setparam *) ev.cmd.data;
+		gii_xwin_cmddata_setparam data;
+
+		GGIDPRINT_MODE("X (setmode_fixed): tell inputlib about new window\n");
+
 		ev.cmd.size = sizeof(gii_cmd_event);
 		ev.cmd.type = evCommand;
 		ev.cmd.target = priv->inp->origin;
 		ev.cmd.code = GII_CMDCODE_XWINSETPARAM;
-		data->win = priv->win;
-		if (data->win == None) data->win = priv->parentwin;
-		data->ptralwaysrel = 0;
+		data.win = priv->win;
+		if (data.win == None) {
+			data.win = priv->parentwin;
+		}
+		data.ptralwaysrel = 0;
+		data.parentwin = priv->parentwin;
+
+		/* Assure aligned memory access. Some platforms
+		 * (i.e. NetBSD/sparc64) rely on this.
+		 */
+		memcpy(ev.cmd.data, &data, sizeof(gii_xwin_cmddata_setparam));
 
 		giiEventSend(priv->inp, &ev);
 	}
 
+	GGIDPRINT_MODE("X (setmode_fixed): set dirty region\n");
+
 	/* ggiOpen will dirty the whole screen for us by calling fillscreen */
 	priv->dirtytl.x = 1; priv->dirtybr.x = 0;
 
 	if (priv->opmansync) MANSYNC_cont(vis);
 
+	GGIDPRINT_MODE("X (setmode_fixed): return code = %i\n", err);
 	return err;
 
 err1:
 	priv->freefb(vis);
 err0:
+	if (priv->opmansync) MANSYNC_cont(vis);
 	return err;
 }
 
@@ -609,8 +759,9 @@ int _ggi_x_resize(ggi_visual_t vis, int w, int h, ggi_event *ev)
 	GGIDPRINT_DRAW("_ggi_x_resize(%p, %dx%d, %p) called\n",
 		       vis, w, h, ev);
 
-	if (LIBGGI_MODE(vis)->visible.x == w &&
-	    LIBGGI_MODE(vis)->visible.y == h) {
+	if ((LIBGGI_X(vis) == w) &&
+	    (LIBGGI_Y(vis) == h))
+	{
 		return 1;
 	}
 
