@@ -1,4 +1,4 @@
-/* $Id: visual.c,v 1.9 2002/09/08 21:37:45 soyt Exp $
+/* $Id$
 ******************************************************************************
 
    Display-FBDEV: visual handling
@@ -111,13 +111,14 @@ static const gg_option optlist[] =
 #define MAX_DEV_LEN	63
 #define DEFAULT_FBNUM	0
 
-extern int GGI_fbdev_resetmode(ggi_visual *vis);
+extern int GGI_fbdev_mode_reset(ggi_visual *vis);
+extern void GGI_fbdev_color_reset(ggi_visual *vis);
 
 static void
 switchreq(void *arg)
 {
 	ggi_visual *vis = arg;
-	ggi_fbdev_priv *priv = LIBGGI_PRIVATE(vis);
+	ggi_fbdev_priv *priv = FBDEV_PRIV(vis);
 	gii_event ev;
 	ggi_cmddata_switchrequest *data;
 
@@ -142,7 +143,7 @@ static void
 switching(void *arg)
 {
 	ggi_visual *vis = arg;
-	ggi_fbdev_priv *priv = LIBGGI_PRIVATE(vis);
+	ggi_fbdev_priv *priv = FBDEV_PRIV(vis);
 
 	GGIDPRINT_MISC("display-fbdev: switching(%p) called\n", vis);
 
@@ -157,42 +158,6 @@ switching(void *arg)
 #endif
 }
 
-
-static void
-do_setpalette(ggi_visual *vis)
-{
-	ggi_graphtype gt = LIBGGI_MODE(vis)->graphtype;
-	ggi_fbdev_priv *priv = LIBGGI_PRIVATE(vis);
-	ggi_color *colormap;
-	struct fb_cmap cmap;
-	int len, i;
-
-	if (GT_SCHEME(gt) != GT_PALETTE) return;
-
-	len = 1 << GT_DEPTH(gt);
-	cmap.start  = 0;
-	cmap.len    = len;
-	cmap.red    = priv->reds;
-	cmap.green  = priv->greens;
-	cmap.blue   = priv->blues;
-	cmap.transp = NULL;
-	colormap = vis->palette;
-
-	for (i = 0; i < len; i++) {
-		priv->reds[i]   = colormap[i].r;
-		priv->greens[i] = colormap[i].g;
-		priv->blues[i]  = colormap[i].b;
-	}
-
-	if (fbdev_doioctl(vis, FBIOPUTCMAP, &cmap) < 0) {
-		GGIDPRINT_COLOR("display-fbdev: PUTCMAP failed.");
-	} else {
-		GGIDPRINT_COLOR("display-fbdev: restored palette for %p\n",
-				vis);
-	}
-}
-
-
 static void
 switchback(void *arg)
 {
@@ -219,7 +184,15 @@ switchback(void *arg)
 		fbdev_doioctl(vis, FBIOPUT_VSCREENINFO, &priv->var);
 	}
 #endif
-	do_setpalette(vis);
+
+	/* See notes about palette member reuse in color.c */
+	if ((LIBGGI_PAL(vis)->setPalette != NULL) && (vis->opcolor->setpalvec != NULL))
+		vis->opcolor->setpalvec(vis, 0, 1 << GT_DEPTH(LIBGGI_GT(vis)), 
+					LIBGGI_PAL(vis)->clut.data);
+	else if (vis->opcolor->setgammamap != NULL)
+		vis->opcolor->setgammamap(vis, 0, vis->gamma->len, 
+					  LIBGGI_PAL(vis)->clut.data);
+
 	if (priv->fix.xpanstep != 0 || priv->fix.ypanstep != 0) {
 		fbdev_doioctl(vis, FBIOPAN_DISPLAY, &priv->var);
 	}
@@ -230,7 +203,7 @@ switchback(void *arg)
 static int 
 GGI_fbdev_sendevent(ggi_visual *vis, gii_event *ev)
 {
-	ggi_fbdev_priv *priv = LIBGGI_PRIVATE(vis);
+	ggi_fbdev_priv *priv = FBDEV_PRIV(vis);
 
 	GGIDPRINT_MISC("GGI_fbdev_sendevent() called\n");
 
@@ -291,72 +264,9 @@ GGI_fbdev_idleaccel(ggi_visual *vis)
 	return 0;
 }
 
-
-static void 
-save_palette(ggi_visual *vis)
-{
-	ggi_fbdev_priv *priv = LIBGGI_PRIVATE(vis);
-	struct fb_cmap cmap;
-	int len;
-
-	if ((priv->orig_var.bits_per_pixel < 1) || 
-	    (priv->orig_var.bits_per_pixel > 8)) {
-		priv->orig_reds = NULL;
-		return;
-	}
-	len = 1 << priv->orig_var.bits_per_pixel;
-	
-	priv->orig_reds = malloc(sizeof(uint16)*len*3);
-	if (priv->orig_reds == NULL) {
-		return;
-	}
-
-	cmap.start = 0;
-	cmap.len   = len;
-	cmap.red   = priv->orig_reds;
-	cmap.green = priv->orig_greens = priv->orig_reds + len;
-	cmap.blue  = priv->orig_blues  = priv->orig_greens + len;
-	cmap.transp = NULL;
-
-	if (ioctl(LIBGGI_FD(vis), FBIOGETCMAP, &cmap) < 0) {
-		GGIDPRINT_COLOR("display-fbdev: GETCMAP failed.\n");
-	} else {
-		GGIDPRINT_COLOR("display-fbdev: Saved palette (len=%d).\n",
-				cmap.len);
-	}
-}
-
-
-static void 
-restore_palette(ggi_visual *vis)
-{
-	ggi_fbdev_priv *priv = LIBGGI_PRIVATE(vis);
-	struct fb_cmap cmap;
-
-	if (priv->orig_reds == NULL) return;
-	
-	cmap.start  = 0;
-	cmap.len    = 1 << priv->orig_var.bits_per_pixel;
-	cmap.red    = priv->orig_reds;
-	cmap.green  = priv->orig_greens;
-	cmap.blue   = priv->orig_blues;
-	cmap.transp = NULL;
-
-	if (ioctl(LIBGGI_FD(vis), FBIOPUTCMAP, &cmap) < 0) {
-		GGIDPRINT_COLOR("display-fbdev: PUTCMAP failed.\n");
-	} else {
-		GGIDPRINT_COLOR("display-fbdev: Restored palette (len=%d).\n", 
-				cmap.len);
-	}
-
-	free(priv->orig_reds);
-	priv->orig_reds = NULL;
-}
-
-
 static int do_cleanup(ggi_visual *vis)
 {
-	ggi_fbdev_priv *priv = LIBGGI_PRIVATE(vis);
+	ggi_fbdev_priv *priv = FBDEV_PRIV(vis);
 	ggi_fbdev_timing *curtim;
 
 	/* We may be called more than once due to the LibGG cleanup stuff */
@@ -370,9 +280,8 @@ static int do_cleanup(ggi_visual *vis)
 		ioctl(LIBGGI_FD(vis), FBIOPUT_CON2FBMAP, &origconmap);
 #endif
 #endif
-		GGI_fbdev_resetmode(vis);
-		restore_palette(vis);
-		close(LIBGGI_FD(vis));
+		GGI_fbdev_color_reset(vis);
+		GGI_fbdev_mode_reset(vis);
 	}
 
 	if (vis->input != NULL) {
@@ -438,10 +347,18 @@ get_fbdev(void)
 
 	/* Find a framebuffer to open */
 	for (fb=0; fb < 32; fb++) {
+#ifdef HAVE_SNPRINTF
+		snprintf(devname, MAX_DEV_LEN, "/dev/fb%d", fb);
+#else
 		sprintf(devname, "/dev/fb%d", fb);
+#endif
 		fd = open(devname, O_RDONLY);
 		if (fd >= 0) break;
+#ifdef HAVE_SNPRINTF
+		snprintf(devname, MAX_DEV_LEN, "/dev/fb/%d", fb);
+#else
 		sprintf(devname, "/dev/fb/%d", fb);
+#endif
 		fd = open(devname, O_RDONLY);
 		if (fd >= 0) break;
 	}
@@ -660,18 +577,17 @@ static int GGIopen(ggi_visual *vis, struct ggi_dlhandle *dlh,
 		}
 	}
 
-	LIBGGI_PRIVATE(vis) = priv = malloc(sizeof(ggi_fbdev_priv));
+	priv = calloc(1, sizeof(ggi_fbdev_priv));
 	if (priv == NULL) {
 		return GGI_ENOMEM;
 	}
 	GGIDPRINT("display-fbdev: Got private mem.\n");
-	
+	LIBGGI_PRIVATE(vis) = priv;
+
 	priv->fb_ptr = NULL;
 	priv->need_timings = 1;
 	priv->flags = 0;
 
-	priv->orig_reds = NULL;
-
 	priv->dohalt = 1;
 	priv->autoswitch = 1;
 	priv->switchpending = 0;
@@ -688,16 +604,14 @@ static int GGIopen(ggi_visual *vis, struct ggi_dlhandle *dlh,
 	if (strlen(options[OPT_DEV].result)) {
 		devfile = options[OPT_DEV].result;
 	} else if (getenv("FRAMEBUFFER") != NULL) {
-		strncpy(devicename, getenv("FRAMEBUFFER"), MAX_DEV_LEN);
-		/* Make sure string is terminated */
-		devicename[MAX_DEV_LEN] = '\0';
+		ggstrlcpy(devicename, getenv("FRAMEBUFFER"), sizeof(devicename));
 		devfile = devicename;
 	} else {
 		fbnum = get_fbdev();
 	}
 
 	GGIDPRINT("display-fbdev: Parsing input options.\n");
-	
+
 	priv->inputs = FBDEV_INP_KBD | FBDEV_INP_MOUSE;
 
 	if (toupper(options[OPT_NOKBD].result[0]) != 'N') {
@@ -717,7 +631,7 @@ static int GGIopen(ggi_visual *vis, struct ggi_dlhandle *dlh,
 	GGIDPRINT("display-fbdev: Parsing physz options.\n");
 	do {
 		int err;
-		err = _ggi_parse_physz(options[OPT_PHYSZ].result, 
+		err = _ggi_physz_parse_option(options[OPT_PHYSZ].result, 
 				       &(priv->physzflags), &(priv->physz)); 
 		if (err != GGI_OK) {
 			do_cleanup(vis);
@@ -899,7 +813,6 @@ static int GGIopen(ggi_visual *vis, struct ggi_dlhandle *dlh,
 		do_cleanup(vis);
 		return GGI_ENODEVICE;
 	}
-	save_palette(vis);
 
 	LIBGGI_GC(vis) = priv->normalgc = malloc(sizeof(ggi_gc));
 	if (priv->normalgc == NULL) {
@@ -907,12 +820,23 @@ static int GGIopen(ggi_visual *vis, struct ggi_dlhandle *dlh,
 		return GGI_ENOMEM;
 	}
 
-	if (get_timings(priv, modedb) == GGI_ENOMEM) {
+	switch (get_timings(priv, modedb)) {
+	case GGI_OK:
+		break;
+	case GGI_ENOFILE:
+		/* PDA's like Compaq iPaq and Sharp Zaurus don't
+		 * need /etc/fb.modes to work, so we ignore this failure.
+		 */
+		break;
+	case GGI_ENOMEM:
 		/* If we can't do a simple malloc() it's better to fail
-		   right now. */
+		 * right now.
+		 */
 		do_cleanup(vis);
 		return GGI_ENOMEM;
-	}
+	default:
+		break;
+	}	/* switch */
 
 	/* Mode management */
 	vis->opdisplay->getmode   = GGI_fbdev_getmode;
@@ -940,6 +864,9 @@ static int GGIclose(ggi_visual *vis, struct ggi_dlhandle *dlh)
 }
 
 
+EXPORTFUNC
+int GGIdl_fbdev(int func, void **funcptr);
+
 int GGIdl_fbdev(int func, void **funcptr)
 {
 	switch (func) {
